@@ -40,6 +40,16 @@
 
 /* ----------------------------- stack ring code --------------------------- */
 
+/* If more than this many transients are in a single branch of a transient
+ * tree, they will end up in more or less random stacking order. */
+#define MAX_TRANSIENTS_IN_BRANCH 200000
+/* Same for total levels of transients. */
+#define MAX_TRANSIENT_LEVELS     10000
+/* This number must fit in a signed int! */
+#define LOWER_PENALTY (MAX_TRANSIENTS_IN_BRANCH * MAX_TRANSIENT_LEVELS)
+
+static void __RaiseOrLowerWindow(
+  FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window);
 static void RaiseOrLowerWindow(
   FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window);
 #if 0
@@ -47,6 +57,8 @@ static void ResyncFvwmStackRing(void);
 #endif
 static void ResyncXStackingOrder(void);
 static void BroadcastRestack(FvwmWindow *s1, FvwmWindow *s2);
+static int collect_transients_recursive(
+  FvwmWindow *t, FvwmWindow *list_head, int layer, Bool do_lower);
 
 #define DEBUG_STACK_RING 1
 #ifdef DEBUG_STACK_RING
@@ -377,10 +389,8 @@ static void raise_over_unmanaged(FvwmWindow *t)
 
 
 static Bool must_move_transients(
-  FvwmWindow *t, Bool do_lower, Bool *found_transient)
+  FvwmWindow *t, Bool do_lower)
 {
-  *found_transient = False;
-
   if (IS_ICONIFIED(t))
   {
     return False;
@@ -392,18 +402,14 @@ static Bool must_move_transients(
     Bool scanning_above_window = True;
     FvwmWindow *q;
 
-    for (q = Scr.FvwmRoot.stack_next; q != &Scr.FvwmRoot; q = q->stack_next)
+    for (q = Scr.FvwmRoot.stack_next;
+	 q != &Scr.FvwmRoot && t->layer <= q->layer; q = q->stack_next)
     {
       if (t->layer < q->layer)
       {
 	/* We're not interested in higher layers. */
 	continue;
       }
-      if (t->layer > q->layer)
-      {
-	/* We are at the end of this layer. Stop scanning windows. */
-	break;
-      }
       else if (t == q)
       {
 	/* We found our window. All further transients are below it. */
@@ -411,22 +417,12 @@ static Bool must_move_transients(
       }
       else if (IS_TRANSIENT(q) && (q->transientfor == t->w))
       {
-	/* found a transient */
-	*found_transient = True;
-
-        /* transients are not allowed below main in the same layer */
-        if ( !scanning_above_window )
-        {
-          return True;
-        }
+	return True;
       }
-      else if ((scanning_above_window && !do_lower) ||
-	       (*found_transient && do_lower))
+      else if (scanning_above_window && !do_lower)
       {
 	/* raise: The window is not raised, so itself and all transients will
 	 * be raised. */
-	/* lower: There is a transient above some other window, so we have to
-	 * lower. */
 	return True;
       }
     }
@@ -435,7 +431,7 @@ static Bool must_move_transients(
 }
 
 static void restack_windows(
-  FvwmWindow *r, FvwmWindow *s, int count, Bool do_broadcast_all)
+  FvwmWindow *r, FvwmWindow *s, int count, Bool do_broadcast_all, Bool do_lower)
 {
   FvwmWindow *t;
   unsigned int flags;
@@ -473,15 +469,15 @@ static void restack_windows(
     }
   }
 
-  changes.sibling = s->frame;
+  changes.sibling = (do_lower) ? r->frame : s->frame;
   if (changes.sibling != None)
   {
-    changes.stack_mode = Above;
+    changes.stack_mode = (do_lower) ? Below : Above;
     flags = CWSibling|CWStackMode;
   }
   else
   {
-    changes.stack_mode = Below;
+    changes.stack_mode = (do_lower) ? Above : Below;
     flags = CWStackMode;
   }
   XConfigureWindow (dpy, r->stack_next->frame, flags, &changes);
@@ -502,22 +498,147 @@ static void restack_windows(
   return;
 }
 
-static void RaiseOrLowerWindow(
+static void __sort_transient_ring(FvwmWindow *ring)
+{
+  FvwmWindow *s;
+  FvwmWindow *t;
+  FvwmWindow *u;
+  FvwmWindow *prev;
+
+  if (ring->stack_next->stack_next == ring)
+  {
+    /* only one or zero windows */
+    return;
+  }
+  /* Implementation note:  this sorting algorithm is about the most inefficient
+   * possible.  It just swaps the position of two adjacent windows in the ring
+   * if they are in the wrong order.  Since transient windows are rare, this
+   * should not cause any notable performance hit.  Because it is important that
+   * the order of windows with the same key is not changed, we can not just use
+   * qsort() here. */
+  for (t = ring->stack_next, prev = ring; t->stack_next != ring;
+       prev = t->stack_prev)
+  {
+    s = t->stack_next;
+    if (t->scratch.i < s->scratch.i)
+    {
+      /* swap windows */
+      u = s->stack_next;
+      s->stack_next = t;
+      t->stack_next = u;
+      u = t->stack_prev;
+      t->stack_prev = s;
+      s->stack_prev = u;
+      s->stack_prev->stack_next = s;
+      t->stack_next->stack_prev = t;
+      if (prev != ring)
+      {
+	/* move further up the ring? */
+	t = prev;
+      }
+      else
+      {
+	/* hit start of ring */
+      }
+    }
+    else
+    {
+      /* correct order, advance one window */
+      t = t->stack_next;
+    }
+  }
+
+  return;
+}
+
+static Bool __restack_window(
+  FvwmWindow *t, Bool do_lower, Bool do_restack_transients, Bool is_new_window)
+{
+  FvwmWindow *s, *r, tmp_r;
+  int count;
+  int test_layer;
+
+  /* detach t, so it doesn't make trouble in the loops */
+  remove_window_from_stack_ring(t);
+
+  count = 1;
+  count += get_visible_icon_window_count(t);
+
+  if (do_restack_transients)
+  {
+    /* collect the transients in a temp list */
+    tmp_r.stack_prev = &tmp_r;
+    tmp_r.stack_next = &tmp_r;
+    count = collect_transients_recursive(t, &tmp_r, t->layer, do_lower);
+    if (count == 0)
+    {
+      do_restack_transients = False;
+    }
+    count++;
+  }
+
+  test_layer = t->layer;
+  if (do_lower)
+  {
+    test_layer--;
+  }
+  /* now find the place to reinsert t and friends */
+  for (s = Scr.FvwmRoot.stack_next; s != &Scr.FvwmRoot; s = s->stack_next)
+  {
+    if (test_layer >= s->layer)
+    {
+      break;
+    }
+  }
+  r = s->stack_prev;
+
+  if (do_restack_transients)
+  {
+    /* re-sort the transient windows according to their scratch.i register */
+    __sort_transient_ring(&tmp_r);
+    /* insert all transients between r and s. */
+    add_windowlist_to_stack_ring_after(&tmp_r, r);
+  }
+
+  /*
+  ** Re-insert t - below transients
+  */
+  add_window_to_stack_ring_after(t, s->stack_prev);
+
+  if (is_new_window && IS_TRANSIENT(t) && DO_STACK_TRANSIENT_PARENT(t) &&
+      !IS_ICONIFIED(t))
+  {
+    /* now that the new transient is properly positioned in the stack ring,
+     * raise/lower it again so that its parent is raised/lowered too */
+    RaiseOrLowerWindow(t, do_lower, True, False);
+    /* make sure the stacking order is correct - may be the sledge-hammer
+     * method, but the recursion ist too hard to understand. */
+    ResyncXStackingOrder();
+    return True;
+  }
+  else
+  {
+    /* restack the windows between r and s */
+    restack_windows(r, s, count, do_restack_transients, do_lower);
+  }
+
+  return False;
+}
+
+/* Important note:  __RaiseOrLowerWindow must *only* be called by
+ * RaiseOrLowerWindow()! */
+static void __RaiseOrLowerWindow(
   FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window)
 {
-  FvwmWindow *s, *r, *t2, *next, tmp_r;
-  int count;
+  FvwmWindow *t2;
   Bool do_move_transients;
-  Bool found_transient;
-  Bool no_movement;
-  int test_layer;
 
   /* Do not raise this window after command execution (see HandleButtonPress()).
    */
   SET_SCHEDULED_FOR_RAISE(t, 0);
 
   /* New windows are simply raised/lowered without touching the transientfor
-   * at first.  Then, further down in the code, RaiseOrLowerWindow() is called
+   * at first.  Then, further down in the code, __RaiseOrLowerWindow() is called
    * again to raise/lower the transientfor if necessary.  We can not do the
    * recursion stuff for new windows because the must_move_transients() call
    * needs a properly ordered stack ring - but the new window is still at the
@@ -542,14 +663,27 @@ static void RaiseOrLowerWindow(
       {
         if (t2->w == t->transientfor)
         {
-	  if (IS_ICONIFIED(t2))
+	  if (IS_ICONIFIED(t2) || t->layer != t2->layer)
 	  {
 	    break;
 	  }
+	  if (do_lower && (!IS_TRANSIENT(t2) || !DO_STACK_TRANSIENT_PARENT(t2)))
+	  {
+	    /* hit the highest level transient; lower this subtree below all
+	     * other subtrees of the same window */
+	    t->scratch.i = -LOWER_PENALTY;
+	  }
+	  else
+	  {
+	    /* Add a bonus to the stack ring position for this branch of the
+	     * transient tree over all other branches. */
+	    t->scratch.i = MAX_TRANSIENTS_IN_BRANCH;
+	  }
+	  __RaiseOrLowerWindow(t2, do_lower, True, False);
           if ((!do_lower && DO_RAISE_TRANSIENT(t2)) ||
               (do_lower && DO_LOWER_TRANSIENT(t2))  )
           {
-	    RaiseOrLowerWindow(t2,do_lower,False,False);
+	    /* moving the parent moves our window already */
 	    return;
           }
         }
@@ -559,103 +693,16 @@ static void RaiseOrLowerWindow(
 
   if (is_new_window)
   {
-    no_movement = False;
     do_move_transients = False;
   }
   else
   {
-    do_move_transients = must_move_transients(t, do_lower, &found_transient);
-
-    /*
-     * This part implements (Dont)FlipTransient style.
-     *
-     * must_move_transients() believes that main should always be below
-     * its transients. Therefore if it finds a transient but does not wish
-     * move it, this means main and its superior transients are already
-     * in the correct position at the top or bottom of the layer - depending
-     * on do_lower.
-     */
-    no_movement = found_transient && !do_move_transients;
+    do_move_transients = must_move_transients(t, do_lower);
   }
-
-  if (!no_movement)
+  if (__restack_window(
+	t, do_lower, do_move_transients, is_new_window) == True)
   {
-    /* detach t, so it doesn't make trouble in the loops */
-    remove_window_from_stack_ring(t);
-
-    count = 1;
-    count += get_visible_icon_window_count(t);
-
-    if (do_move_transients)
-    {
-      /* collect the transients in a temp list */
-      tmp_r.stack_prev = &tmp_r;
-      tmp_r.stack_next = &tmp_r;
-      for (t2 = Scr.FvwmRoot.stack_next; t2 != &Scr.FvwmRoot; t2 = next)
-      {
-	next = t2->stack_next;
-	if ((IS_TRANSIENT(t2)) && (t2->transientfor == t->w) &&
-	    (t2->layer == t->layer) && !IS_ICONIFIED(t2))
-	{
-	  /* t2 is a transient to lower */
-	  count++;
-	  count += get_visible_icon_window_count(t2);
-
-	  /* unplug it */
-	  remove_window_from_stack_ring(t2);
-	  /* put it above tmp_r */
-	  add_window_to_stack_ring_after(t2, tmp_r.stack_prev);
-	}
-      }
-      if (tmp_r.stack_next == &tmp_r)
-      {
-	do_move_transients = False;
-      }
-    }
-
-    test_layer = t->layer;
-    if (do_lower)
-    {
-      test_layer--;
-    }
-    /* now find the place to reinsert t and friends */
-    for (s = Scr.FvwmRoot.stack_next; s != &Scr.FvwmRoot; s = s->stack_next)
-    {
-      if (test_layer >= s->layer)
-      {
-	break;
-      }
-    }
-    r = s->stack_prev;
-
-    if (do_move_transients && tmp_r.stack_next != &tmp_r)
-    {
-      /* insert all transients between r and s. */
-      add_windowlist_to_stack_ring_after(&tmp_r, r);
-    }
-
-    /*
-    ** Re-insert t - below transients
-    */
-    add_window_to_stack_ring_after(t, s->stack_prev);
-
-    if (is_new_window && IS_TRANSIENT(t) && DO_STACK_TRANSIENT_PARENT(t) &&
-	!IS_ICONIFIED(t))
-    {
-      /* now that the new transient is properly positioned in the stack ring,
-       * raise/lower it again so that its parent is raised/lowered too */
-      RaiseOrLowerWindow(t, do_lower, True, False);
-      /* make sure the stacking order is correct - may be the sledge-hammer
-       * method, but the recursion ist too hard to understand. */
-      ResyncXStackingOrder();
-      return;
-    }
-    else
-    {
-      /* restack the windows between r and s */
-      restack_windows(
-	r, s, count, (do_move_transients && tmp_r.stack_next != &tmp_r));
-    }
+    return;
   }
 
   if (!do_lower)
@@ -669,7 +716,6 @@ static void RaiseOrLowerWindow(
       raise_over_unmanaged(t);
     }
 
-
     /*
      * The following is a hack to raise X windows over native windows
      * which is needed for some (all ?) X servers running under Windows
@@ -701,18 +747,31 @@ static void RaiseOrLowerWindow(
         XRaiseWindow (dpy, t2->frame);
       }
     }
-
-
     /*  This needs to be done after all the raise hacks.  */
     raisePanFrames();
-  }
-  /* If the window has been raised, make sure the decorations are updated
-   * immediately in case we are in a complex function (e.g. raise, unshade). */
-  if (!do_lower)
-  {
+    /* If the window has been raised, make sure the decorations are updated
+     * immediately in case we are in a complex function (e.g. raise, unshade).
+     */
     XSync(dpy, 0);
     handle_all_expose();
   }
+
+  return;
+}
+
+static void RaiseOrLowerWindow(
+  FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window)
+{
+  FvwmWindow *fw;
+
+  /* clean the auxiliary registers used in stacking transients */
+  for (fw = Scr.FvwmRoot.next; fw != NULL; fw = fw->next)
+  {
+    fw->scratch.i = 0;
+  }
+  __RaiseOrLowerWindow(t, do_lower, allow_recursion, is_new_window);
+
+  return;
 }
 
 /*
@@ -1107,11 +1166,11 @@ void mark_transient_subtree(
     SET_IN_TRANSIENT_SUBTREE(s, 0);
     if (IS_TRANSIENT(s) && (layer < 0 || layer == s->layer))
     {
-      s->pscratch = get_transientfor_fvwmwindow(s);
+      s->scratch.p = get_transientfor_fvwmwindow(s);
     }
     else
     {
-      s->pscratch = NULL;
+      s->scratch.p = NULL;
     }
   }
   /* now loop over the windows and mark the ones we need to move */
@@ -1130,7 +1189,7 @@ void mark_transient_subtree(
 
       if (IS_IN_TRANSIENT_SUBTREE(s))
 	continue;
-      if (DO_ICONIFY_WINDOW_GROUPS(s) && s->wmhints &&
+      if (use_window_group_hint && DO_ICONIFY_WINDOW_GROUPS(s) && s->wmhints &&
 	  (s->wmhints->flags & WindowGroupHint) &&
 	  (s->wmhints->window_group != None) &&
 	  (s->wmhints->window_group != s->w) &&
@@ -1140,7 +1199,7 @@ void mark_transient_subtree(
       }
       if (!IS_TRANSIENT(s) && !use_group_hint)
 	continue;
-      r = (FvwmWindow *)s->pscratch;
+      r = (FvwmWindow *)s->scratch.p;
       if (do_ignore_icons && r && IS_ICONIFIED(r))
 	continue;
       if (IS_TRANSIENT(s))
@@ -1152,6 +1211,8 @@ void mark_transient_subtree(
 	{
 	  /* have to move this one too */
 	  SET_IN_TRANSIENT_SUBTREE(s, 1);
+	  /* used for stacking transients */
+	  s->scratch.i += r->scratch.i + 1;
 	  /* need another scan through the list */
 	  is_finished = False;
 	  continue;
@@ -1216,6 +1277,7 @@ void new_layer(FvwmWindow *tmp_win, int layer)
   FvwmWindow list_head;
   int add_after_layer;
   int count;
+  Bool do_lower;
 
   tmp_win = get_transientfor_top_fvwmwindow(tmp_win);
   if (layer == tmp_win->layer)
@@ -1235,11 +1297,13 @@ void new_layer(FvwmWindow *tmp_win, int layer)
   {
     /* lower below the windows in the new (lower) layer */
     add_after_layer = layer;
+    do_lower = True;
   }
   else
   {
     /* raise above the windows in the new (higher) layer */
     add_after_layer = layer + 1;
+    do_lower = False;
   }
   /* find the place to insert the windows */
   for (target = Scr.FvwmRoot.stack_next; target != &Scr.FvwmRoot;
@@ -1261,7 +1325,8 @@ void new_layer(FvwmWindow *tmp_win, int layer)
     GNOME_SetLayer(tmp_win);
   }
   /* move the windows without modifying their stacking order */
-  restack_windows(list_head.stack_next->stack_prev, target, count, (count > 1));
+  restack_windows(
+    list_head.stack_next->stack_prev, target, count, (count > 1), do_lower);
 
   return;
 }
@@ -1287,6 +1352,15 @@ static Bool is_on_top_of_layer_ignore_rom(FvwmWindow *fw)
   FvwmWindow *t;
   Bool ontop = True;
 
+  if (IS_SCHEDULED_FOR_DESTROY(fw))
+  {
+    /* stack ring members are no longer valid */
+    return False;
+  }
+  if (DO_RAISE_TRANSIENT(fw))
+  {
+    mark_transient_subtree(fw, fw->layer, MARK_RAISE, True, False);
+  }
   for (t = fw->stack_prev; t != &Scr.FvwmRoot; t = t->stack_prev)
   {
     if (t->layer > fw->layer)
@@ -1302,8 +1376,7 @@ static Bool is_on_top_of_layer_ignore_rom(FvwmWindow *fw)
      * windows, but on top of its layer, it would be considered on top. */
     if (Scr.bo.RaiseOverUnmanaged || overlap(fw, t))
     {
-      if (!IS_TRANSIENT(t) || t->transientfor != fw->w ||
-	  !DO_RAISE_TRANSIENT(fw))
+      if (!DO_RAISE_TRANSIENT(fw) || (!IS_IN_TRANSIENT_SUBTREE(t) && t != fw))
       {
 	ontop = False;
 	break;
