@@ -18,6 +18,7 @@
 
 typedef struct
 {
+  unsigned do_broadcast_focus : 1;
   unsigned do_redecorate : 1;
   unsigned do_redecorate_transient : 1;
   unsigned do_redraw_decoration : 1;
@@ -28,6 +29,7 @@ typedef struct
   unsigned do_update_icon : 1;
   unsigned do_update_icon_boxes : 1;
   unsigned do_update_icon_font : 1;
+  unsigned do_update_icon_placement : 1;
   unsigned do_update_icon_title : 1;
   unsigned do_update_mini_icon : 1;
   unsigned do_update_stick : 1;
@@ -40,8 +42,8 @@ typedef struct
   unsigned do_update_modules_flags : 1;
 } update_win;
 
+void destroy_scheduled_windows(void);
 void apply_decor_change(FvwmWindow *tmp_win);
 void flush_window_updates(void);
-void update_styles_func(F_CMD_ARGS);
 
 #endif /* _UPDATE_ */
